"""
support/seano/views/shared/hlist.py

Infrastructure to work with hierarchical lists (hlists)
"""
from functools import reduce

from .markup import (
    SUPPORTED_MARKUP,
    SeanoMarkup,
    SeanoFalseyMarkup,
    SeanoMarkupException,
)


class SeanoUnlocalizedHListNode(object):
    """
    Represents a hierarchical list of ``SeanoMarkup`` objects, commonly referred
    to in the ``seano`` ecosystem as an "hlist".

    Main features:

    - some automatic type coercion
    - some error checking
    - developer-friendly iteration utilities
    """
    def __init__(self, element, children):
        self.element = element
        self.children = children

        # `children` is supposed to be a `list` of `SeanoUnlocalizedHListNode`
        # objects that each have a `SeanoMarkup` object as their `element`.

        # To simplify code elsewhere, allow the top-level `children` container
        # to be some limited variants, such as None, generators, or markup:
        if self.children is None:
            self.children = [] # no children
        elif isinstance(self.children, SeanoUnlocalizedHListNode):
            pass # correct type (nothing to change)
        elif isinstance(self.children, SeanoMarkup):
            self.children = [self.children]
        elif hasattr(self.children, '__iter__'): # allow generators
            pass # close enough (list comprehension below will convert to list)
        else:
            raise SeanoMarkupException('assertion error: unsupported hlist children list: %s' % (self.children,))

        # To simplify code elsewhere, allow each item inside `children` to be
        # some limited variants, such as markup:
        def fixup_type(item):
            if item is None:
                raise SeanoMarkupException('assertion error: no hlist node child is allowed to be None')
            elif isinstance(item, SeanoUnlocalizedHListNode) and not item.element:
                raise SeanoMarkupException('assertion error: no hlist node child is allowed to have a falsey element; proposed node child: %s' % (item,))
            elif isinstance(item, SeanoUnlocalizedHListNode):
                return item
            elif isinstance(item, SeanoFalseyMarkup):
                return SeanoUnlocalizedHListNode(element=item, children=None) # Special case
            elif isinstance(item, SeanoMarkup) and not item:
                raise SeanoMarkupException('assertion error: no hlist node child is allowed to have a falsey element; proposed node child: %s' % (item,))
            elif isinstance(item, SeanoMarkup):
                return SeanoUnlocalizedHListNode(element=item, children=None)
            raise SeanoMarkupException('assertion error unsupported hlist child payload: %s' % (item,))
        self.children = [fixup_type(x) for x in self.children]

    def deep_copy(self):
        return SeanoUnlocalizedHListNode(
            element=None if self.element is None else self.element.deep_copy(),
            children=[x.deep_copy() for x in self.children],
        )

    def __bool__(self): return bool(self.element) or bool(self.children)
    def __nonzero__(self): return bool(self.element) or bool(self.children)

    def __eq__(self, other): return isinstance(other, SeanoUnlocalizedHListNode) and self.element == other.element and self.children == other.children

    def __str__(self): return '%s -> %s' % (self.element, self.children)
    def __repr__(self): return str(self)

    def walk(self, level=1, include_root=False, unroll_recursion=True):
        """
        Recursively iterates over this object, providing a custom iteration
        node object at each step, with special named properties describing
        the current state of navigation.  Notably:

        - level: the integer level of the current node in the hlist; starts at 1
        - level_diff: the change in the level since the last node
        - text: the text payload of this node
        - num_children: the number of children of this node
        - note_ids: the list of note IDs that contributed to this node

        The values provided during iteration are deep clones of this object,
        which means that you are free to mutate them during iteration, and
        that any custom values you set on this object will be missing.
        """
        tree = SeanoUnlocalizedHListIterNode(node=self, level=level - 1)
        for node in tree._walk(include_root=include_root, unroll_recursion=unroll_recursion):
            yield node

    def __iter__(self):
        """
        A shortcut for the default behavior of the ``walk()`` method.
        """
        for node in self.walk():
            yield node

    def first(self):
        """
        Returns just the first element, if it exists.  Returns a
        ``SeanoFalseyMarkup`` object if this hlist is empty.

        This method is more performant than going through the iterator API,
        because the iterator API always does a full traversal of the entire
        tree before it returns the first node.
        """
        for child in self.children:
            return child.element
        return SeanoFalseyMarkup('You asked for the first element in a SeanoUnlocalizedHListNode object, but there were no contents at all')

    def __add__(self, other):
        result = self.deep_copy()
        result.merge(other)
        return result

    def merge(self, other):
        for incoming in other.children:
            for check in self.children:
                if check.element.payload == incoming.element.payload:
                    check.element.tags = sorted(set(check.element.tags).union(incoming.element.tags))
                    check.merge(incoming)
                    break
            else:
                self.children.append(incoming.deep_copy())
        return self


class SeanoUnlocalizedHListIterNode(SeanoUnlocalizedHListNode):
    """
    A subclass of ``SeanoUnlocalizedHListNode`` that is designed for more
    convenient iteration.

    Main features:

    - Can stand-in as a normal ``SeanoUnlocalizedHListNode`` object
    - When initialized, performs a deep clone of the original, so that you
      can mutate this copy without mutating the original
    - Cloning this class intentionally returns a ``SeanoUnlocalizedHListNode``
      tree, not a tree of this class
    - Adds some conveneince properties, such as `level`, `num_children`,
      and `note_ids`
    """
    def __init__(self, node, level):
        # Because `node` is a `SeanoUnlocalizedHListNode`, we can assume it
        # has already been sanitized, which means we can skip the expensive
        # sanitizing that `SeanoUnlocalizedHListNode` performs:
        self.element = node.element
        self.children = [__class__(x, level + 1) for x in node.children]
        # Adding in additional behavior:
        self.level = level
        self.num_children = len(node.children)
        self.note_ids = getattr(node.element, 'tags', []) # Not all markup is real (i.e., `SeanoFalseyMarkup`)

    def _walk(self, include_root=False, unroll_recursion=True):
        if include_root:
            yield self
        for child in self.children:
            yield child
            if unroll_recursion:
                for grandchild in child._walk():
                    yield grandchild


def _parse_hlist_node(data, localization, note_id, markup_constructor):
    if isinstance(data, list):
        # either list or hlist (don't care which yet)
        for v in data:
            for result in _parse_hlist_node(data=v, localization=localization, note_id=note_id, markup_constructor=markup_constructor):
                yield result
        return

    if isinstance(data, dict):
        # hlist (convert key/value to element/children)
        for k, v in data.items():
            yield SeanoUnlocalizedHListNode(
                element=markup_constructor(payload=k, localization=localization, tags=[note_id] if note_id else []),
                children=list(_parse_hlist_node(data=v, localization=localization, note_id=note_id, markup_constructor=markup_constructor))
            )
        return

    # all other values are assumed to be markup text, though sometimes Yaml
    # can decide to use other types...  pass the raw value to the markup
    # constructor and hope it works.
    yield SeanoUnlocalizedHListNode(
        element=markup_constructor(payload=data, localization=localization, tags=[note_id] if note_id else []),
        children=None,
    )

def _parse_list_loc_hlist(data, pick_locs, note_id, markup_constructor):
    for item in data:
        for loc in pick_locs:
            if loc in item:
                for result in _parse_hlist_node(
                    data=item[loc],
                    localization=loc,
                    note_id=note_id,
                    markup_constructor=markup_constructor,
                ):
                    yield result
                break
        else:
            yield SeanoFalseyMarkup('Missing localization')


def _parse_loc_hlist(data, pick_locs, note_id, markup_constructor):
    for loc in pick_locs:
        if loc in data:
            return _parse_hlist_node(
                data=data[loc],
                localization=loc,
                note_id=note_id,
                markup_constructor=markup_constructor,
            )
    return SeanoFalseyMarkup('Missing localization')


def _parse_unloc_hlist(data, pick_locs, note_id, markup_constructor):
    return _parse_hlist_node(
        data=data,
        localization=None,
        note_id=note_id,
        markup_constructor=markup_constructor,
    )


SUPPORTED_STRUCTURES = [
    ('-list-loc',   _parse_list_loc_hlist),
    ('-loc-list',   _parse_loc_hlist),
    ('-loc-hlist',  _parse_loc_hlist),
    ('-loc',        _parse_loc_hlist),
    ('',            _parse_unloc_hlist),
]

def seano_get_struct_parser(key):
    key_root, _, markup_language = key.rpartition('-')
    try:
        markup_constructor = SUPPORTED_MARKUP[markup_language]
    except KeyError:
        key_root = key
        markup_constructor = SUPPORTED_MARKUP['']  # plain text

    for suffix, constructor in SUPPORTED_STRUCTURES:
        if key_root.endswith(suffix):
            return lambda **kw: constructor(
                markup_constructor=markup_constructor,
                **kw
            )

    raise SeanoMarkupException('internal error')


def seano_read_hlist(notes, keys, localizations):
    """
    A big, fat, one-stop-shop for reading keys from multiple note objects,
    merging them, enabling you to iterate over them, and encapsulating the
    type of markup used for each fragment.

    Dirty little secret: under the hood, most structures we type into `seano`
    notes are convertable into hlists, even if they aren't officially hlists.
    For example, `*-loc-hlist-*`, `*-list-loc-*`, and `*-loc-*` are all
    readable by this function.  Yes, somewhat confusingly, if you use this
    function to read `example-loc-md`, this function will return an hlist with
    one element, despite that field being a simple localized blob.  In the
    long term, this actually turns out to be useful because it lets you read
    that field from more than one note (the `notes` parameter is a list).
    This behavior also conveniently permits hlist values in fields that expect
    blobs (a developer convenience when writing documentation), without
    forcing the view layer to have two different paths for rendering blobs.

    IMPORTANT: This function is not designed to read a field that is *not*
    some kind of container of rich text markup.

    This function returns a `SeanoUnlocalizedHListNode`` object, which is a
    recursive container that represents the entire hlist, and provides
    developer-friendly iteration tools.

    Due to the schema we've chosen to use in `seano` notes, this function is
    your last chance to choose the desired localization.  The result returned
    from this function is always un-localized, though it does contain memory
    of the localization it came from, for use in embedding in the output
    document.

    `keys` and `localizations` are both treated as priority lists.  The first
    match in any given case will be used.

    `notes` is a list of note objects in which to search.  All of the notes are
    always searched.
    """
    def _inner():
        for note in notes:
            for key in keys:
                if key in note:
                    for hunk in seano_get_struct_parser(key)(
                        note_id   = note.get('id') or None,
                        data      = note[key],
                        pick_locs = localizations,
                    ):
                        if hunk.element:
                            yield SeanoUnlocalizedHListNode(element=None, children=[hunk])
                        else:
                            yield hunk
                    break

    return reduce(lambda a, b: a.merge(b), _inner(), SeanoUnlocalizedHListNode(element=None, children=None))



"""
support/seano/views/shared/links.py

Infrastructure to help process links to external servers
"""
import re

github_issue_url_regex = re.compile(r'^https?://github.com/[^/]+/([^/]+)/issues/([0-9]+)$')
jira_url_regex = re.compile(r'^https?://[^/]*jira[^/]*/browse/([A-Z]+\-[0-9]+)$')
redmine_url_regex = re.compile(r'^https?://[^/]*redmine[^/]*/issues/([0-9]+)$')


def get_ticket_display_name(url):
    '''
    Returns the display name of the ticket addressable at the given URL.
    '''
    m = github_issue_url_regex.match(url)
    if m: return '/'.join([m.group(1), m.group(2)])

    m = jira_url_regex.match(url)
    if m: return m.group(1)

    m = redmine_url_regex.match(url)
    if m: return m.group(1)

    raise Exception("Don't know how to parse the given ticket URL: %s" % (url,))



"""
support/seano/views/shared/metacache.py

Infrastructure to help with efficient caching of metadata related to seano schemas
"""
import json


class SeanoMetaCache(object):
    """
    A centralized, portable, transparent caching mechanism for metadata related to seano query data.

    The general concept is that wherever you want a "releases list", use this class instead, and
    functions can set members on this object for caching without you needing to realize that they're
    doing it, or need to manage caching on your own.

    For context, a number of release ancestry analysis functions require deep traversals of the
    release ancestry graph.  Because the release ancestry graph is not large, this isn't terribly
    expensive on the face of it...  until you realize that some views of data perform deep traversals
    for each release, and sometimes multiple times over a range of other fields...  Once you start
    doing deep graph traversals N^2 or N^3 times (and a deep graph traversal itself has N^N
    complexity on its own!), those times begin to add up quickly.

    **WARNING**: Be careful about mutating data saved to this class.  Because random functions cache
    data on this object *(that is the purpose of this class, after all)*, it is conceptually possible
    to make the caches obsolete without properly invalidating them.  If you desire to mutate the data
    on this object, you should generally prefer to make a whole new ``SeanoMetaCache`` object (which
    would have nothing cached), and use that copy to store the new data.  This helps ensure that
    caches stored on a ``SeanoMetaCache`` object are always valid.
    """
    def __init__(self, serialized_seano_query_output_data):
        """
        Initializes a ``SeanoMetaCache`` object.

        Parameters:

        - ``serialized_seano_query_output_data`` (string): The raw data from a seano query output

        This function deserializes the given seano query output data, grabs the releases object, and
        maps the releases list to some commonly used schemas.  When this initializer finishes, the
        following members are set on this class:

        - ``everything``: The top-level dictionary in the seano query output file
        - ``releases``: The original releases list inside the seano query output file
        - ``named_releases``: A copy of the releases list, converted to a dictionary, keyed by the
          names of each release.
        """
        self.everything = json.loads(serialized_seano_query_output_data)
        self.releases = self.everything['releases']
        self.named_releases = {release['name']: release for release in self.releases}



"""
support/seano/views/shared/text_buf.py

Infrastructure to help build text blobs
"""
import sys
import tempfile


def to_ascii(txt):
    # ABK: Pylint can't read this if statement, and complains about `unicode` not existing in Python 3.
    if sys.hexversion >= 0x3000000:
        if isinstance(txt, bytes):
            return txt
        if isinstance(txt, str):
            return txt.encode('utf-8')
        return str(txt).encode('utf-8')
    if isinstance(txt, str):
        return txt
    if isinstance(txt, unicode): #pylint: disable=E0602
        return txt.encode('utf-8')
    return str(txt)


def to_unicode(txt):
    # ABK: Pylint can't read this if statement, and complains about `unicode` not existing in Python 3.
    if sys.hexversion >= 0x3000000:
        if isinstance(txt, bytes):
            return txt.decode('utf-8')
        if isinstance(txt, str):
            return txt
        return str(txt)
    if isinstance(txt, str):
        return txt.decode('utf-8')
    if isinstance(txt, unicode): #pylint: disable=E0602
        return txt
    return unicode(txt) #pylint: disable=E0602


class FencedTextBuffer(object):
    '''
    Helps you write a text blob with an optional prefix and suffix.  Designed to help you write the
    prefix and suffix up-front, while the code that builds the body may be long.  This helps keep
    the prefix and suffix close to each other when they are declared.

    You should use this class within a ``with`` statement so that proper cleanup happens if an
    exception is raised.

    This class accepts both ASCII strings UTF-8 strings, for both Python 2 and Python 3.

    The udump() method returns data in Unicode, regardless of the Python version.
    '''
    def __init__(self, prefix=None, suffix=None, skip_fences_when_body_empty=False):
        'Creates a default ``FencedTextBuffer`` object.'
        self.prefix = prefix
        self.body = tempfile.TemporaryFile()
        self.suffix = suffix
        self.skip_fences_when_body_empty = skip_fences_when_body_empty

    def __enter__(self):
        '''
        For compatibility with ``with``.  We have nothing special to do ourselves, but we should
        forward this notice to each of the file buffers.
        '''
        self.body.__enter__()
        return self

    def __exit__(self, ertype, value, traceback):
        '''
        For compatibility with ``with``.  We have nothing special to do ourselves, but we should
        forward this notice to each of the file buffers.
        '''
        self.body.__exit__(ertype, value, traceback)

    def write(self, txt):
        '''
        Writes the given ASCII or UTF-8 data to the body.
        '''
        self.body.write(to_ascii(txt))

    def udump(self, insert_before_body=None, insert_after_body=None):
        '''
        Returns the finalized, concatenated string version of all data currently saved in this
        object.

        Optionally accepts text to insert before or after the body hunk during assembly.  As far
        as the ``skip_fences_when_body_empty`` option is concerned, these injected values are
        "part of" (they add to) the body.

        The returned data is ALWAYS a unicode string, regardless of Python version.  This means
        it's a unicode object in Python 2, and a str object in Python 3.
        '''
        self.body.seek(0, 0)

        hunk = [insert_before_body, self.body.read(), insert_after_body]

        if not self.skip_fences_when_body_empty or any(hunk):
            hunk = [self.prefix] + hunk + [self.suffix]

        return u''.join(map(to_unicode, filter(None, hunk)))



"""
support/seano/views/shared/html_buf.py

Infrastructure to help build HTML files
"""
from .text_buf import FencedTextBuffer

try:
    from html import escape as _html_escape_func # correct on python 3; explodes on python 2
except ImportError:
    # Must be python 2
    from cgi import escape as _html_escape_func

def html_escape(txt):
    # ABK: So, pylint doesn't follow the try/except logic above, which results in it issuing W1505 saying that
    #      `cgi.escape` is deprecated **ON PYTHON 3**.  In my experimentation, this is the easiest workaround.
    return _html_escape_func(txt) #pylint: disable=W1505


class SeanoHtmlFragment(object):
    '''
    Encapsulates an HTML fragment, in a world where the HTML fragment may require that certain CSS or JS be added in
    the headder of the final document.  A "fragmented fragment", if you will.
    '''
    def __init__(self, html, css=None, js=None): #pylint: disable=C0103
        self.html = html or ''
        self.css = css or ''
        self.js = js or '' #pylint: disable=C0103

    def __bool__(self):
        if not self.html and not self.css and not self.js:
            return False
        return True
    __nonzero__ = __bool__


class SeanoHtmlBuffer(object):
    '''
    Helps you write a single-file HTML file, in a world where:

    - it's a fairly non-trivial document, so you want to use a file buffer for performance
    - you don't know all of the CSS and JS prior to writing the body
    - you want to support Unicode & Python 2 & Python 3 (file streams are a pain)

    You should use this class within a ``with`` statement so that proper cleanup happens if an exception is raised.

    This class accepts both ASCII strings UTF-8 strings.

    This class does not HTML-encode anything.  As far as the caller is concerned, the only thing this class does
    automatically is auto-translate between ASCII and UTF-8.  If something has to be HTML-encoded (such as most JS!),
    the caller must do it prior to writing the data here.
    '''
    def __init__(self):
        '''
        Creates a default ``SeanoHtmlBuffer`` object.  Default data includes:

        - HTML, HEAD, and BODY elements
        - UTF-8 character encoding declaration
        - Mobile-friendly viewport declaration
        '''
        self.doc = FencedTextBuffer(prefix='<html>', suffix='</html>')
        self.head = FencedTextBuffer(prefix='<head>', suffix='</head>', skip_fences_when_body_empty=True)
        self.css = FencedTextBuffer(prefix='<style type="text/css">', suffix='</style>', skip_fences_when_body_empty=True) #pylint: disable=C0301
        self.js = FencedTextBuffer(prefix='<script>', suffix='</script>', skip_fences_when_body_empty=True) #pylint: disable=C0103
        self.body = FencedTextBuffer(prefix='<body>', suffix='</body>')

        self.write_head('''<meta charset="utf-8">''')
        self.write_head('''<meta name="viewport" content="width=device-width, initial-scale=1.0">''')

    def __enter__(self):
        '''
        For compatibility with ``with``.  We have nothing special to do ourselves, but we should forward this notice to
        each of the file buffers.
        '''
        self.doc.__enter__()
        self.head.__enter__()
        self.css.__enter__()
        self.js.__enter__()
        self.body.__enter__()
        return self

    def __exit__(self, ertype, value, traceback):
        '''
        For compatibility with ``with``.  We have nothing special to do ourselves, but we should forward this notice to
        each of the file buffers.
        '''
        self.doc.__exit__(ertype, value, traceback)
        self.head.__exit__(ertype, value, traceback)
        self.css.__exit__(ertype, value, traceback)
        self.js.__exit__(ertype, value, traceback)
        self.body.__exit__(ertype, value, traceback)

    def write_head(self, txt):
        '''
        Writes the given ASCII or UTF-8 data to the generic ``<HEAD>`` section.
        '''
        self.head.write(txt)

    def write_css(self, txt):
        '''
        Writes the given ASCII or UTF-8 data to the CSS section.
        '''
        self.css.write(txt)

    def write_js(self, txt):
        '''
        Writes the given ASCII or UTF-8 data to the JavaScript section.
        '''
        self.js.write(txt)

    def write_body(self, txt):
        '''
        Writes the given ASCII or UTF-8 data to the ``<BODY>`` section.
        '''
        self.body.write(txt)

    def all_data(self):
        '''
        Returns the finalized, concatenated string version of all data currently saved in this object.

        The returned data is ALWAYS a unicode string, regardless of Python version.  This means it's a unicode object
        in Python 2, and a str object in Python 3.
        '''
        return self.doc.udump(
            insert_after_body=self.head.udump(
                insert_after_body=self.css.udump() + self.js.udump()
            ) + self.body.udump()
        )



"""
support/seano/views/qa_notes.py

Infrastructure to convert a seano query output file into what is known as the QA Notes page (single-file HTML+CSS+JS)

The public entry point here is the function named ``compile_qa_notes()``.
"""
import argparse
import datetime
import sys
from .shared.components import *
from .shared.hlist import seano_read_hlist, SeanoUnlocalizedHListNode
from .shared.html_buf import SeanoHtmlBuffer, html_escape as escape
from .shared.markup import SeanoMarkdown, SeanoReStructuredText
from .shared.metacache import SeanoMetaCache
from .shared.schema_plumbing import seano_minimum_descendant_list


class QANotesRenderInfrastructure(object):
    '''
    The render infrastructure has stack-local state.  That means the storage of such state cannot be global, or
    else this infrastructure becomes non-thread-safe (and resetting state between runs becomes more complicated).
    The two other options are to use function-local storage, and class-local storage.  Between those two, class-
    local storage is easier to unit test.  (Not that we have unit tests, but I can dream, right?)
    '''
    def __init__(self):
        self._next_elem_uid = 0

    def compile_ticket_url(self, url):  #pylint: disable=R0201
        if url is None:
            # Visually matches font size of tickets in support/seano/views/shared/components.py, despite
            # this font size being...  larger?  It's probably an illusion caused by use of italics.
            return '<span style="font-size:90%"><em>No ticket associated</em></span>'
        return seano_render_html_ticket(url)

    def _get_elem_uid(self):
        self._next_elem_uid = self._next_elem_uid + 1
        return self._next_elem_uid

    def run(self, srcdata):
        self._next_elem_uid = 0
        cmc = SeanoMetaCache(srcdata)
        with SeanoHtmlBuffer() as f:
            f.write_head('''<title>QA Notes for ''')
            f.write_head(escape(cmc.everything['project_name']['en-US']))
            f.write_head(''' v''')
            f.write_head(escape(cmc.releases[0]['name']))
            f.write_head('''</title>''')
            f.write_css(SeanoReStructuredText('sample text').toHtmlBlock().css) # Write out Pygments' CSS sheet
            f.write_css('''
body {
    font-family: sans-serif;
    -webkit-text-size-adjust: 100%;
}
a {
    color: #0175bb; /* Bahama */
}
a:visited {
    color: #1997EB; /* Denim */
}
blockquote {
    border-left: 0.2em solid #8DD2FC; /* Cornflower */
}
pre, pre.code, code, tt.literal {
    background-color: #F4FAFB; /* Lily White */
    border: 0.05em solid #BDE6FE; /* French Pass */
    border-radius: 0.2em;
}
pre {
    font-size: 110%;
    overflow: scroll;
    padding: 0.6em;
    margin-left: 1.5em;
}
code, tt.literal {
    display: inline-block;
    font-family: Courier, monospace;
    padding: 0 0.1em 0 0.2em;
}
pre > code {
    padding: 0;
    border: none;
}
.unimportant-long-sha1 {
    word-break: break-all;
}

.build-uniq-div {
    background: #D3ECD6; /* Pistachio */
    padding: 0.5em;
    margin-bottom: 1em;
}
.build-uniq-div > .head {
    display: block;
    margin: 0 0 0.5em 0;
}
.build-uniq-div > .build-uniq-data {
    background: white;
}
.build-uniq-div > .build-uniq-data > .data {
    display: inline-block;
    margin: 0.5em 1em 0.5em 1em;
}

.release-head {
    background: #8DD2FC; /* Cornflower */
    padding: 0.5em;
    margin-bottom: 0.2em;
}
.release-head > .release-name {
    display: inline-block;
    font-size: 110%;
    font-weight: bold;
}
.release-head > .release-since {
    display: inline-block;
    margin-left: 1em;
}
.release-head > .show-release, .release-head > .hide-release {
    float: right;
}
.release-head > .show-release > a, .release-head > .hide-release > a {
    color: black;
}

.note-head {
    display: inline-block;
    margin: 0 0 1em 0;
}
.note-head > .internal-short {
    font-weight: bold;
}
.note-head > .ticket {
    display: inline-block;
    margin-left: 1em;
    font-size: 90%;
}
.release-subhead > .show-release-notes, .release-subhead > .show-qa-notes {
    display: inline-block;
    margin-left: 1em;
    margin-bottom: 0.1em;
    padding: 0.2em 0.7em 0.2em 0.7em;
}
.release-subhead > .hide-release-notes, .release-subhead > .hide-qa-notes {
    display: inline-block;
    margin-left: 1em;
    margin-bottom: 0.1em;
    padding: 0.2em 0.7em 0.2em 0.7em;
    background: #BDE6FE; /* French Pass */
}
.note-head > .show-technical, .note-head > .hide-technical {
    font-size: 75%; /* Matches font size of tickets in support/seano/views/shared/components.py */
}
.release-notes-body {
    margin: 1em;
    padding: 1em;
    background: rgb(236,236,236); /* Off-white of background of System Preferences in light mode */
}
.rnhover {
    background: #BDE6FE; /* French Pass */
}
.public-release-notes, .internal-release-notes, .testing, .technical {
    margin-left: 1em;
}
.custsrv-release-notes {
    margin-left: 2em;
}
.custsrv-release-notes > :first-child {
    margin-left: -1em;
}
p {
    margin-top: 0;
}
ol, ul {
    margin-bottom: 1em;
}
.clarification {
    font-weight: normal;
    font-size: 80%;
    font-style: italic;
    color: grey;
}
@media (prefers-color-scheme: dark) {
    body {
        background-color: #292A2F; /* Xcode's off-black background color */
        color: white;
    }
    a {
        color: #8DD2FC; /* Cornflower */
    }
    a:visited {
        color: #E5F5FE; /* Aqua Spring */
    }
    blockquote {
        border-left: 0.2em solid #0175bb; /* Bahama */
    }
    pre, pre.code, code, tt.literal {
        background-color: #454545; /* Custom dark gray */
        border: 0.1em solid #07466D; /* Regal Blue */
        border-radius: 0.3em;
    }
    pre > code {
        border: none;
    }
    .build-uniq-div, .build-uniq-div > .head {
        background: #38823E; /* Goblin */
    }
    .build-uniq-div > .build-uniq-data {
        background: black;
    }
    .release-head,
    .release-head > .release-name,
    .release-head > .release-since,
    .release-head > .show-release > a,
    .release-head > .hide-release > a {
        background-color: #0175bb; /* Bahama */
    }
    .release-head > .show-release > a, .release-head > .hide-release > a {
        color: white;
    }
    .release-subhead > .hide-release-notes, .release-subhead > .hide-qa-notes {
        background: #8DD2FC; /* Cornflower */
    }
    .release-subhead > .hide-release-notes > a, .release-subhead > .hide-qa-notes > a {
        color: black;
        background: #8DD2FC; /* Cornflower */
    }
    .release-notes-body {
        background-color: rgb(63,65,68); /* Cocoa window title bar color in dark mode */
    }
    .rnhover {
        background: #0175bb; /* Bahama */
    }
}''')
            f.write_js('''function showRelease(id) {
    document.getElementById('show-release-' + id).style.display = 'none';
    document.getElementById('hide-release-' + id).style.display = 'inline-block';
    document.getElementById('release-body-' + id).style.display = 'block';
}
function hideRelease(id) {
    document.getElementById('show-release-' + id).style.display = 'inline-block';
    document.getElementById('hide-release-' + id).style.display = 'none';
    document.getElementById('release-body-' + id).style.display = 'none';
}
function showReleaseNotes(id) {
    document.getElementById('show-release-notes-' + id).style.display = 'none';
    document.getElementById('hide-release-notes-' + id).style.display = 'inline-block';
    document.getElementById('release-notes-' + id).style.display = 'block';
}
function hideReleaseNotes(id) {
    document.getElementById('show-release-notes-' + id).style.display = 'inline-block';
    document.getElementById('hide-release-notes-' + id).style.display = 'none';
    document.getElementById('release-notes-' + id).style.display = 'none';
}
function showQaNotes(id) {
    document.getElementById('show-qa-notes-' + id).style.display = 'none';
    document.getElementById('hide-qa-notes-' + id).style.display = 'inline-block';
    document.getElementById('qa-notes-' + id).style.display = 'block';
}
function hideQaNotes(id) {
    document.getElementById('show-qa-notes-' + id).style.display = 'inline-block';
    document.getElementById('hide-qa-notes-' + id).style.display = 'none';
    document.getElementById('qa-notes-' + id).style.display = 'none';
}
function showTechnical(id) {
    document.getElementById('show-technical-' + id).style.display = 'none';
    document.getElementById('hide-technical-' + id).style.display = 'inline-block';
    document.getElementById('technical-' + id).style.display = 'block';
}
function hideTechnical(id) {
    document.getElementById('show-technical-' + id).style.display = 'inline-block';
    document.getElementById('hide-technical-' + id).style.display = 'none';
    document.getElementById('technical-' + id).style.display = 'none';
}''')
            f.write_body('''<h2>QA Notes for ''')
            f.write_body(escape(cmc.everything['project_name']['en-US']))
            f.write_body(' v')
            f.write_body(escape(cmc.releases[0]['name']))
            f.write_body('</h2><p>Commit <code class="unimportant-long-sha1">')
            # IMPROVE: ABK: This lies when the working directory is dirty.  How best to fix?
            f.write_body(escape(cmc.releases[0].get('commit', None)                         # Normal committed work
                                or (len(cmc.releases) > 1 and cmc.releases[1].get('commit', None)) # Dirty checkout
                                or '???'))                                                       # Unexpected error
            f.write_body('</code>, built on ')
            # IMPROVE: ABK: This line makes the output non-deterministic.  Is that bad?
            # IMPROVE: ABK: Yes, that is bad.  The date should be present in the query output.
            f.write_body(datetime.datetime.today().strftime('%m/%d/%Y at %I:%M %p %Z'))
            f.write_body('</p>')
            bu = seano_read_hlist([cmc.everything], ['build-uniqueness-list-md', 'build-uniqueness-list-rst'], ['en-US'])
            if bu:
                f.write_body('<div class="build-uniq-div"><span class="head">Build Uniqueness</span>')
                f.write_body('<div class="build-uniq-data">')
                for data in sorted([x.element.toHtmlLine().html for x in bu], key=lambda s: s.lower()):
                    f.write_body('<span class="data">')
                    f.write_body(data)
                    f.write_body('</span>')
                f.write_body('</div></div>')
            release_count = 0
            for release in cmc.releases:
                release_count = release_count + 1
                if release_count > 5:
                    break
                self.write_release(f, release, cmc, release_count <= 1)

            return f.all_data()

    def write_release(self, f, release, cmc, is_first_release):
        release_div_id = self.write_release_head(f, release, cmc, is_first_release)
        self.write_release_body(f, release, is_first_release, release_div_id)

    def write_release_head(self, f, release, cmc, is_first_release):
        release_div_id = self._get_elem_uid()
        f.write_body('<div class="release-head"><span class="release-name">Changes in ')
        f.write_body(escape(release['name']))
        f.write_body('</span><span class="release-since">(since ')
        bag = [x['name'] for x in release['after']]
        bag = seano_minimum_descendant_list(bag, cmc)
        f.write_body(escape(' and '.join(bag) or 'the dawn of time'))
        f.write_body(')</span>')
        f.write_body('<span class="show-release" id="show-release-%d" style="display:%s">' \
                     '''<a href="javascript:showRelease('%d')">Show</a></span>''' % (
                         release_div_id, 'inline-block' if not is_first_release else 'none', release_div_id))
        f.write_body('<span class="hide-release" id="hide-release-%d" style="display:%s">' \
                     '''<a href="javascript:hideRelease('%d')">Hide</a></span>''' % (
                         release_div_id, 'none' if not is_first_release else 'inline-block', release_div_id))
        f.write_body('</div>') # end of "release-head" div
        return release_div_id

    def write_release_body(self, f, release, is_first_release, release_body_div_id):
        f.write_body('<div id="release-body-%d" class="release-body" style="display:%s">' % (
                         release_body_div_id, 'none' if not is_first_release else 'block'))
        release_notes_id, qa_notes_id = self.write_release_section_toggles(f)
        self.write_release_notes(f, release, release_notes_id)
        self.write_qa_notes(f, release, qa_notes_id)
        f.write_body('</div>')

    def write_release_section_toggles(self, f):
        release_notes_id = self._get_elem_uid()
        qa_notes_id = self._get_elem_uid()
        f.write_body('<div class="release-subhead">')
        f.write_body('<span class="show-release-notes" id="show-release-notes-%d" style="display:inline-block">' \
                     '''<a href="javascript:showReleaseNotes('%d')">Release Notes</a></span>''' % (
                         release_notes_id, release_notes_id))
        f.write_body('<span class="hide-release-notes" id="hide-release-notes-%d" style="display:none">' \
                     '''<a href="javascript:hideReleaseNotes('%d')">Release Notes</a></span>''' % (
                         release_notes_id, release_notes_id))
        f.write_body('<span class="show-qa-notes" id="show-qa-notes-%d" style="display:none">' \
                     '''<a href="javascript:showQaNotes('%d')">QA Notes</a></span>''' % (
                         qa_notes_id, qa_notes_id))
        f.write_body('<span class="hide-qa-notes" id="hide-qa-notes-%d" style="display:inline-block">' \
                     '''<a href="javascript:hideQaNotes('%d')">QA Notes</a></span>''' % (
                         qa_notes_id, qa_notes_id))
        f.write_body('</div>') # end of "release-subhead" div
        return release_notes_id, qa_notes_id

    def write_release_notes(self, f, release, release_notes_id): #pylint: disable=R0201
        def render_mouse_hover_toggle_logic(identifiers):
            # Here, we expect that we are within the attribute list of the start of an HTML element of some kind.
            # Here, we add the class, onmouseover, and onmouseleave elements.
            # The caller is expected to close this element (i.e., write the '>')
            return ''.join([
                ' class="',
                ' '.join(identifiers),
                '" onmouseover="',
                ';'.join(['''Array.prototype.forEach.call(document.getElementsByClassName('%s'), function(e){e.classList.toggle('rnhover', true)})''' % (s,) for s in identifiers]), #pylint: disable=C0301
                '" onmouseleave="',
                ';'.join(['''Array.prototype.forEach.call(document.getElementsByClassName('%s'), function(e){e.classList.toggle('rnhover', false)})''' % (s,) for s in identifiers]), #pylint: disable=C0301
                '"',
            ])

        def write_hlist(hlist, default, is_blob_field=False, include_tickets=False):
            if not hlist:
                f.write_body(default)
                return
            def block_formatter(node):
                base_func = seano_html_hlist_blob_formatter_simple
                result = base_func(node=node)
                styles = ['r%dp%s' % (release_notes_id, id) for id in node.note_ids]
                return '<div' + render_mouse_hover_toggle_logic(styles) + '>' + result + '</div>'
            def line_formatter(node):
                base_func = seano_html_hlist_line_formatter_text_with_tickets(notes=release.get('notes') or []) if include_tickets \
                    else seano_html_hlist_line_formatter_simple
                result = base_func(node=node)
                styles = ['r%dp%s' % (release_notes_id, id) for id in node.element.tags]
                return '<span' + render_mouse_hover_toggle_logic(styles) + '>' + result + '</span>'
            f.write_body(seano_render_html_hlist(hlist, is_blob_field=is_blob_field, block_formatter=block_formatter, line_formatter=line_formatter))
        f.write_body('<div id="release-notes-%d" class="release-notes-body" style="display:none">' %(release_notes_id,))
        f.write_body('<div class="public-release-notes">')
        backstory_clarification = ''
        if any([x.get('is-copied-from-backstory') for x in release.get('notes') or []]):
            backstory_clarification = ' <span class="clarification">(includes backstories)</span>'
        f.write_body('<h4>Public Release Notes' + backstory_clarification + '</h4>')
        write_hlist(hlist=seano_read_hlist(notes=release.get('notes') or [],
                                          keys=['customer-short-loc-hlist-md', 'customer-short-loc-hlist-rst'],
                                          localizations=['en-US']),
                    default='<p><em>No public release notes</em></p>')
        f.write_body('</div>')
        f.write_body('<div class="internal-release-notes">')
        f.write_body('<h4>Internal Release Notes</h4>')
        write_hlist(hlist=seano_read_hlist(notes=[x for x in release.get('notes') or [] if not x.get('is-copied-from-backstory')],
                                          keys=['employee-short-loc-hlist-md', 'employee-short-loc-hlist-rst'],
                                          localizations=['en-US']),
                    default='<p><em>No internal release notes</em></p>',
                    include_tickets=True)
        f.write_body('</div>')
        f.write_body('<div class="custsrv-release-notes">')
        f.write_body('<h4>Member Care Notes</h4>')
        write_hlist(hlist=seano_read_hlist(notes=[x for x in release.get('notes') or [] if not x.get('is-copied-from-backstory')],
                                          keys=['mc-technical-loc-md', 'mc-technical-loc-rst'],
                                          localizations=['en-US']),
                    is_blob_field=True,
                    default='<p><em>No Member Care notes</em></p>')
        f.write_body('</div>')
        f.write_body('</div>')

    def write_qa_notes(self, f, release, qa_notes_id):
        f.write_body('<div id="qa-notes-%d">' % (qa_notes_id,))
        notes_new_in_this_release = [x for x in release['notes'] if not x.get('is-copied-from-backstory')]
        if not release['notes']:
            f.write_body('<p class="testing"><em>No changes</em></p>')
        elif not notes_new_in_this_release:
            f.write_body('<p class="testing"><em>No changes, however some release notes were repeated '
                         'from earlier work</em></p>')
        else:
            f.write_body('<ul>')
            for note in notes_new_in_this_release:
                f.write_body('<li><span class="note-head"><span class="internal-short">')
                head = seano_read_hlist([note], ['employee-short-loc-hlist-md', 'employee-short-loc-hlist-rst'], ['en-US']).first()
                f.write_body(head.toHtmlLine().html or 'Internal release note missing')
                f.write_body('</span>') # employee-short-loc-hlist-*
                for t in note.get('tickets', None) or [None]: # None is used to indicate secret work
                    f.write_body('<span class="ticket">')
                    f.write_body(self.compile_ticket_url(t))
                    f.write_body('</span>')
                technical = seano_read_hlist([note], ['employee-technical-loc-md', 'employee-technical-loc-rst'], ['en-US'])
                if technical:
                    tech_id = self._get_elem_uid()
                    f.write_body('<span class="ticket show-technical" id="show-technical-%d">' \
                                 '''<a href="javascript:showTechnical('%d')">More details</a></span>''' % (
                                     tech_id, tech_id))
                    f.write_body('<span class="ticket hide-technical" id="hide-technical-%d" style="display:none">' \
                                 '''<a href="javascript:hideTechnical('%d')">Fewer details</a></span>''' % (
                                     tech_id, tech_id))
                f.write_body('</span>') # note-head
                if technical:
                    f.write_body('<div class="technical" id="technical-%d" style="display:none">' % (tech_id,))
                    f.write_body(seano_render_html_hlist(technical, is_blob_field=True))
                    f.write_body('</div>')
                f.write_body('<div class="testing">')
                f.write_body(seano_render_html_hlist(seano_read_hlist([note], ['qa-technical-loc-md', 'qa-technical-loc-rst'], ['en-US']), is_blob_field=True)
                             or SeanoMarkdown('*QA Notes missing*').toHtmlBlock().html)
                f.write_body('</div></li>')
            f.write_body('</ul>')
        f.write_body('</div>')


def format_qa_notes(*args):
    '''
    Given a Json blob (in serialized form), return the contents of the corresponding QA Notes page.

    The QA Notes page is implemented using HTML+CSS+JS; as such, if you are going to save it to a file, you probably
    want to use a ``.html`` extension.
    '''
    parser = argparse.ArgumentParser()
    parser.prog = ' '.join([parser.prog, 'format', 'qa_notes'])
    parser.add_argument('--src', action='store', default='-', help='Input file; use a single hyphen for stdin; default is to use stdin')
    parser.add_argument('--out', action='store', default='-', help='Output file; use a single hyphen for stdout; default is to use stdout')

    ns = parser.parse_args(args)

    if ns.src in ['-']:
        data = sys.stdin.read()
    else:
        with open(ns.src, 'r') as f:
            data = f.read()

    result = QANotesRenderInfrastructure().run(data)

    if ns.out in ['-']:
        sys.stdout.write(result)
    else:
        with open(ns.out, 'wb') as f:
            f.write(result)



"""
support/seano/views/shared/schema_painting.py

Infrastructure to retroactively modify seano's query output file.

Functionality like this is deliberately not part of seano itself to help keep seano's query
output schema as simple as possible.  These functions don't create any new information
that doesn't already exist, but they do make existing information easier to access on-the-fly.
"""
from .schema_plumbing import *


def seano_paint_backstory_releases(cmc, start):
    """
    Iterates over the releases list in the given ``SeanoMetaCache`` (``cmc``) object, starting
    at the given start release name, and progressing to all ancestors.  On every touched release,
    ``is-backstory`` is set to a boolean value indicating whether or not the release is part of
    a backstory relative to the given start release.

    The releases list is edited in-place.

    The start release is assumed to not be a backstory.

    Releases descendant from the start release are not modified; this algorithm only operates
    on ancestors.
    """
    def paint(name, is_backstory):

        cmc.named_releases[name]['is-backstory'] = is_backstory

        ancestors = cmc.named_releases[name]['after']

        for ancestor in [x['name'] for x in ancestors if x.get('is-backstory', False)]:
            paint(ancestor, True)
        for ancestor in [x['name'] for x in ancestors if not x.get('is-backstory', False)]:
            paint(ancestor, is_backstory)

    paint(start, False)


def seano_paint_release_sys_limits(cmc):
    """
    Iterates over the releases list in the given ``SeanoMetaCache`` (``cmc``) object,
    propagating the min and max support OS fields across the entire release ancestry.

    The releases list is edited in-place.
    """
    fields = ['min-supported-os', 'max-supported-os']
    seano_copy_note_fields_to_releases(cmc, fields)
    seano_propagate_sticky_release_fields(cmc, fields)


def seano_paint_release_risk_levels(cmc):
    """
    Iterates over the releases list in the given ``SeanoMetaCache`` (``cmc``) object,
    calculating the aggregate risk level for each release based on data found in notes.

    The releases list is edited in-place.

    This function is conceptually similar to seano_copy_note_fields_to_releases(),
    except that there is a *default* value when the 'risk' key does not exist in
    any note in the release.
    """
    for release in cmc.releases:
        if 'risk' in release:
            continue

        # Because we're iterating over all releases (including backstories),
        # we need to skip notes in each release that are copied from backstories:
        notes = [n for n in release['notes'] if not n.get('is-copied-from-backstory')]

        levels = [x.get('risk') for x in notes]
        for level in ['high', 'medium', 'low']:
            if level in levels:
                release['risk'] = level
                break
        else:
            release['risk'] = '' if any([x != None for x in levels]) else None



"""
support/seano/views/shared/schema_plumbing.py

Low-level infrastructure to retroactively modify seano's query output file.

Functionality like this is deliberately not part of seano itself to help keep seano's query
output schema as simple as possible.  These functions don't create any new information
that doesn't already exist, but they do make existing information easier to access on-the-fly.
"""
import os


class SeanoSchemaPaintingException(Exception):
    pass


def seano_release_ancestor_names_including_self(name, cmc):
    """
    Returns an unordered set of the names of all releases that are ancestors of the given release
    name in the ancestry graph contained inside the given ``SeanoMetaCache`` (``cmc``) object,
    including the given release name itself.

    Parameters:

    - ``name`` (string): The name of the release you're interested in
    - ``cmc`` (``SeanoMetaCache``): A ``SeanoMetaCache`` object containing the release ancestry graph
      that is pertinent to the release name with which you're inquiring

    Returns: a set of release names (a set of strings)
    """
    try:
        cache = cmc.ancestor_release_name_sets_including_self
    except AttributeError:
        cache = {}
        cmc.ancestor_release_name_sets_including_self = cache
    try:
        return cache[name]
    except KeyError:
        result = set([name]).union(*[
            seano_release_ancestor_names_including_self(x['name'], cmc) for x in cmc.named_releases[name]['after']
        ])
        cache[name] = result
        return result


def seano_release_descendant_names_including_self(name, cmc):
    """
    Returns an unordered set of the names of all releases that are descendants of the given release
    name in the ancestry graph contained inside the given ``SeanoMetaCache`` (``cmc``) object,
    including the given release name itself.

    Parameters:

    - ``name`` (string): The name of the release you're interested in
    - ``cmc`` (``SeanoMetaCache``): A ``SeanoMetaCache`` object containing the release ancestry graph
      that is pertinent to the release name with which you're inquiring

    Returns: a set of release names (a set of strings)
    """
    try:
        cache = cmc.descendant_release_name_sets_including_self
    except AttributeError:
        cache = {}
        cmc.descendant_release_name_sets_including_self = cache
    try:
        return cache[name]
    except KeyError:
        result = set([name]).union(*[
            seano_release_descendant_names_including_self(x['name'], cmc) for x in cmc.named_releases[name]['before']
        ])
        cache[name] = result
        return result


def seano_minimum_ancestor_list(bag, cmc):
    """
    Given a bag of release names (S) and a ``SeanoMetaCache`` (``cmc``) object containing a set of
    releases, this function returns a subset of S (M) where no release in M is a descendant of any
    release in M.

    Parameters:

    - ``bag`` (iterable of strings): The source list of release names
    - ``cmc`` (``SeanoMetaCache``): A ``SeanoMetaCache`` containing a set of releases

    Returns: list of release names (list of strings)
    """
    if not isinstance(bag, list):
        bag = list(bag)

    for item in bag:
        if len(bag) < 2:
            # It's no longer possible to remove any elements
            break
        smaller_bag = [x for x in bag if x != item]
        if item in set().union(*[seano_release_descendant_names_including_self(x, cmc) for x in smaller_bag]):
            bag = smaller_bag

    return bag


def seano_minimum_descendant_list(bag, cmc):
    """
    Given a bag of release names (S) and a ``SeanoMetaCache`` (``cmc``) object containing a set of
    releases, this function returns a subset of S (M) where no release in M is an ancestor of any
    release in M.

    Parameters:

    - ``bag`` (iterable of strings): The source list of release names
    - ``cmc`` (``SeanoMetaCache``): A ``SeanoMetaCache`` containing a set of releases

    Returns: list of release names (list of strings)
    """
    if not isinstance(bag, list):
        bag = list(bag)

    for item in bag:
        if len(bag) < 2:
            # It's no longer possible to remove any elements
            break
        smaller_bag = [x for x in bag if x != item]
        if item in set().union(*[seano_release_ancestor_names_including_self(x, cmc) for x in smaller_bag]):
            bag = smaller_bag

    return bag


def seano_field_mergetool_opaque(does_privileged_base_exist, privileged_base, additions):
    """
    A merge tool used by some seano plumbing that performs a merge of an opaque type.

    Inputs:

    - ``does_privileged_base_exist`` (``bool``): whether or not an existing ``privileged_base`` exists
    - ``privileged_base`` (``any``): a value that ``additions`` are getting merged *into*
    - ``additions`` (``[any]``): a list of values to merge together onto the ``privileged_base``

    Returns: (any) the merged result; ``privileged_base`` is not modified

    Nature of the merge algorithm:

    1. If the ``privileged_base`` exists, then it overrides any ``additions`` and is the final answer.
    2. If all the ``additions`` are the same value, then any one of the ``additions`` is the final answer.

    On any error, a ``SeanoSchemaPaintingException`` is raised.
    """
    if does_privileged_base_exist:
        return privileged_base

    if not additions or not all([x == additions[0] for x in additions]):
        raise SeanoSchemaPaintingException('Unable to merge values: %s' % (additions,))

    return additions[0]


def seano_copy_note_fields_to_releases(cmc, fields):
    """
    Iterates over the releases list inside the given ``SeanoMetaCache`` (``cmc``) object, copying
    from notes onto the each associated release the fields and their values identified by the given
    list of fields.

    The given releases list is edited in-place.

    Inputs:

    - ``cmc`` (``SeanoMetaCache``): a ``SeanoMetaCache`` object containing releases
    - ``fields`` (``list`` or ``dict``): a list of fields to copy from each note onto the
      respective releases, or a dictionary of fields to copy, associated with their merge
      tool functions

    Returns: nothing

    On any error, a ``SeanoSchemaPaintingException`` is raised.

    WARNING: This algorithm is designed for **rarely changing value**.  Things you change once
    in a blue moon.  This algorithm **has no reasonable merge tool**, and when you change a value
    more than once in the same release, things explode.
    """
    if isinstance(fields, list):
        fields = {x: seano_field_mergetool_opaque for x in fields}

    # For each release:
    for r in cmc.releases:

        # For each field we're copying from notes to releases:
        for f in fields.keys():

            # Because we're iterating over all releases (including backstories),
            # we need to skip notes in each release that are copied from backstories:
            notes = [n for n in r['notes'] if not n.get('is-copied-from-backstory')]

            # For each note, grab the value for the current field:
            values = [n[f] for n in notes if f in n]
            if values:

                # Merge the new values into the release:
                try:
                    r[f] = fields[f](
                        does_privileged_base_exist=f in r,
                        privileged_base=r.get(f),
                        additions=values,
                    )
                except SeanoSchemaPaintingException as e:
                    # We are not going to swallow this exception; we will let it unwind the stack.
                    # However, we would like to improve the error message before it goes.
                    msg = e.args[0] + '''

This happened because multiple notes within the {release} release
tried to set a different new value for {field},
and seano isn't smart enough to reconcile the differences and save a
provably correct merged value on the {release} release.  To workaround
this problem, you have two main choices:

1. Create a new release in between the two notes that conflict, such that
   each release edits {field} only once
2. Open up seano-config.yaml, and on the {release} release, manually
   set the correctly merged value of {field}'''.format(release=r['name'], field=f)
                    e.args = (msg,) + e.args[1:]
                    raise


def seano_propagate_sticky_release_fields(cmc, fields):
    """
    Iterates over the releases list inside the given ``SeanoMetaCache`` (``cmc``) object, copying
    from ancestor releases to descendant releases the fields and their values identified by the
    given list of fields.

    The given releases list is edited in-place.

    Inputs:

    - ``cmc`` (``SeanoMetaCache``): a ``SeanoMetaCache`` object containing releases
    - ``fields`` (``list`` or ``dict``): a list of fields to copy from one release to the next,
      or a dictionary of fields to copy, associated with their merge tool functions

    Returns: nothing

    On any error, a ``SeanoSchemaPaintingException`` is raised.

    WARNING: This algorithm is designed for **rarely changing value**.  Things you change once
    in a blue moon.  This algorithm **has no reasonable merge tool**, and when you change a value
    in more than one parallel ancestry, when the ancestries eventually merge, things explode.
    """
    if isinstance(fields, list):
        fields = {x: seano_field_mergetool_opaque for x in fields}

    _seen_releases = set()
    def process_release(release):
        if release['name'] in _seen_releases:
            return
        _seen_releases.add(release['name'])

        # Process all parents first:
        for r in release['after']:
            process_release(cmc.named_releases[r['name']])

        # Copy each field from the parent release, one by one:
        for f in fields:
            # List all non-transitive immediate parents:
            values = seano_minimum_descendant_list(bag=[x['name'] for x in release['after']], cmc=cmc)
            # Convert release names to release objects:
            values = [cmc.named_releases[r] for r in values]
            # Fetch the value of the current field from each of the release objects, if set:
            values = [r[f] for r in values if f in r]
            # If this field was set on any parent:
            if values:
                # ... then copy the value to this release:
                # (note that we may have to reconcile multiple values from multiple parent releases)
                try:
                    release[f] = seano_field_mergetool_opaque(
                        does_privileged_base_exist=f in release,
                        privileged_base=release.get(f),
                        additions=values,
                    )
                except SeanoSchemaPaintingException as e:
                    # We are not going to swallow this exception; we will let it unwind the stack.
                    # However, we would like to improve the error message before it goes.
                    msg = e.args[0] + '''

This happened because multiple ancestors of the {release} release
have changed the value of {field} to different values,
and seano isn't smart enough to reconcile the differences and save a
provably correct merged value on the {release} release.  The easiest
way to workaround this problem is to open up seano-config.yaml, and
on the {release} release, manually set the correctly merged value of
{field}.'''.format(release=release['name'], field=f)
                    e.args = (msg,) + e.args[1:]
                    raise

    for r in reversed(cmc.releases): # Not required, but reduces unnecessary recursion
        process_release(r)



"""
support/seano/views/shared/markup.py

Infrastructure to help compile rich text markup

*********************************************
Institutional Knowledge on how DocUtils works
*********************************************

DocUtils has a "Node tree" concept that is responsible for laying out the
structured form of a document.  Confusingly, there are *two* different node
trees: (a) a class hierarchy of different kinds of Nodes, and (b) a document
hierarchy of specific Node subclass objects that contain the data of the
document.

Certain Node classes are subclasses of other Node classes, which helps share
code that serializes content to a certain filetype.

The document tree is a tree of Node objects; each Node object then contains
a fragment of the document, as it existed in the original reStructuredText
document.

In a sense, Nodes are the "common unifying document type" used internally
by DocUtils to represent a document, prior to serializing to the requested
target document type.

Nodes are *not* responsible for serializing into a specific filetype.  The
"Translator" concept is used to convert the tree of node objects into a
serialized document.  A translator takes a single Node object (which presumably
is the root of a Node tree), and serializes it into the document type it owns
based on the class type of the Node object.  There is usually only one
translator per serialized document type (RTF, PDF, man page, etc).

If you want to support new output file types, what you want to define is a new
Translator subclass.

If you want to support a new directive in reStructuredText, you *may* want to
create a new Node subclass, in particular if what the new directive receives
from the user is conceptually a new kind of data.  For data types already
supported, there's no reason you can't use an existing Node subclass.  Either
way, you want to then create a Directive subclass and register it.  This
registration process appears to be global, and may cause compatibility issues
as complexity grows.  Iterate as needed.
"""
import base64
import json
import re
try:
    from StringIO import StringIO # correct on python 2.x; explodes on python 3.x
except ImportError:
    # Must be python 3.x
    from io import StringIO
import sys
# ABK: Why can't pylint import these modules?
import docutils.core #pylint: disable=E0401
import docutils.nodes #pylint: disable=E0401
import docutils.parsers.rst #pylint: disable=E0401
import docutils.writers.html4css1 #pylint: disable=E0401
import html
import markdown
from .html_buf import SeanoHtmlFragment


class SeanoMarkupException(Exception):
    pass


class SeanoMermaidNode(docutils.nodes.General, docutils.nodes.Element):
    '''
    This is a DocUtils Node subclass that represents the data received from
    the user in a reStructuredText document when they use the ``.. mermaid::``
    directive.

    Nodes only declare data fragment types and store original document data;
    they do not declare the directive itself or serialize any data to any
    specific file format.

    Per DocUtils conventions, this class should be named "mermaid" (lowercase).
    However, that makes me (ABK) nervous, so I'm being more verbose for now.
    '''
    pass


class SeanoMermaidDirective(docutils.parsers.rst.Directive):
    '''
    This is a DocUtils Directive subclass that parses a Mermaid directive
    invocation from reStructuredText and converts it into a Node object
    (specifically the Node subclass named ``SeanoMermaidNode``).

    Per DocUtils conventions, this class should be named "Mermaid" (titlecase).
    However, that makes me (ABK) nervous, so I'm being more verbose for now.
    '''
    has_content = True # Mermaid source passed via what reStructuredText calls the "content"
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec = { # Dictionary of options accepted by this directive
        'alt': docutils.parsers.rst.directives.unchanged,
        'min-width': docutils.parsers.rst.directives.unchanged,
        'max-width': docutils.parsers.rst.directives.unchanged,
    }

    def run(self):
        node = SeanoMermaidNode()
        node['code'] = '\n'.join(self.content)
        node['options'] = {}
        if 'alt' in self.options:
            node['alt'] = self.options['alt']
        if 'min-width' in self.options:
            node['min-width'] = self.options['min-width']
        if 'max-width' in self.options:
            node['max-width'] = self.options['max-width']
        return [node]

# Actually register our Mermaid directive.  This is the line that makes the
# syntax ``.. mermaid::`` work in a reStructuredText document.
docutils.parsers.rst.directives.register_directive('mermaid', SeanoMermaidDirective)


_MERMAID_AUTO_INIT_KEY = 0

class SeanoSingleFileHtmlTranslator(docutils.writers.html4css1.HTMLTranslator):
    '''
    This is a DocUtils Translator subclass that serializes a Node tree into
    what we colloquially call "single-file HTML".  It is built upon DocUtils'
    built-in HTML 4 & CSS 1 translator implementation.

    When using this translator, you should use the ``docutils.writers.html4css1.Writer``
    writer.
    '''

    def visit_SeanoMermaidNode(self, node):
        # ABK: Mermaid doesn't properly calculate the size of elements that are
        # not visible.  To workaround, don't compile Mermaid diagrams until they
        # are visible.  To implement this without requiring infrastructure
        # outside of this method, we're going to give every Mermaid diagram its
        # own unique identifier, and then use the `IntersectionObserver` API to
        # detect when the element becomes visible, and when that happens, tell
        # Mermaid to compile that specific diagram.
        global _MERMAID_AUTO_INIT_KEY
        _MERMAID_AUTO_INIT_KEY = _MERMAID_AUTO_INIT_KEY + 1
        key = 'mermaid-%d' % (_MERMAID_AUTO_INIT_KEY,)

        self.body.extend([
            '<pre class="mermaid" id="%s">' % (key,),
            node['code'],
            '</pre>',
            '''<script type="module">''',
                '''import mermaid from 'https://cdn.jsdelivr.net/npm/mermaid@10/dist/mermaid.esm.min.mjs';''',
                '''const prefersDarkMode = window.matchMedia('(prefers-color-scheme: dark)').matches;''',
                '''mermaid.initialize({'startOnLoad': false, 'securityLevel': 'antiscript', 'theme': prefersDarkMode ? 'dark' : 'neutral'});''',
                '''new IntersectionObserver((entries, observer) => {''',
                    '''entries.forEach(entry => {''',
                        '''if (entry.intersectionRatio > 0) {''',
                            '''observer.disconnect();''',
                            '''mermaid.run(%s);''' % (json.dumps({'querySelector': '#' + key}),),
                        '''}''',
                    '''});''',
                '''}).observe(document.getElementById('%s'));''' % (key,),
            '''</script>''',
        ])

    def depart_SeanoMermaidNode(self, node):
        pass


_dl_elem_pattern = re.compile(r'''\s*</?d[ldt](?: [^>]*)?>\s*''', re.MULTILINE)
def _seano_rst_to_some_html(txt, writer_class, translator_class):
    '''
    Compiles the given reStructuredText blob into an HTML snippet, sans the ``<html>`` and ``<body>`` elements.

    Returns a SeanoHtmlFragment object containing the HTML fragment, and also recommended CSS/JS to make it work.

    On soft errors, this function may return an empty ``SeanoHtmlFragment`` object.  This function never returns None.
    '''
    if not txt.strip(): return SeanoHtmlFragment(html='')
    # Docutils likes to print warnings & errors to stderr & the compiled output.  We don't particularly want that
    # here...  What we'd like ideally is to capture any warnings and errors, and explicitly report them to the
    # caller.  If this is being used within Waf, we'd ideally like to trigger a build failure (and never have
    # surprise output in the rendered HTML).
    #
    # Fortunately, Docutils lets us do this.
    #
    # More info on settings overrides here:
    #   https://sourceforge.net/p/docutils/mailman/message/30882883/
    #   https://github.com/pypa/readme_renderer/blob/master/readme_renderer/rst.py
    #
    # Documentation on how to use custom translator objects:
    #   https://gist.github.com/mastbaum/2655700
    #
    error_accumulator = StringIO()
    writer = writer_class()
    writer.translator_class = translator_class
    parts = docutils.core.publish_parts(txt, writer=writer, settings_overrides={
        'warning_stream': error_accumulator,
    })

    # Artificially fail if any errors or warnings were reported
    errors = error_accumulator.getvalue().splitlines()
    error_accumulator.close() # ABK: Not sure if this is required, but most of the docs have it
    if errors:
        # ABK: Some errors include the bad markup, and some don't.  Not sure what the pattern is yet.
        #      For now, for all errors, append the original full markup, with line numbers.
        with_line_numbers = []
        for line in txt.splitlines():
            with_line_numbers.append('%4d    %s' % (len(with_line_numbers) + 1, line))
        errors.append('    %s' % ('\n    '.join(with_line_numbers)))
        raise SeanoMarkupException('\n'.join(errors))

    # No errors; return the rendered HTML fragment
    html = parts['fragment']
    css = parts['stylesheet']

    # Docutils likes to insert <dl>, <dd>, and <dt> elements.  Long term, it would be nice to know why (screen readers
    # come to mind).  For now, those elements are causing problems with styling.  Yank them out.
    html = _dl_elem_pattern.sub('', html)

    # The CSS that Docutils returns is wrapped inside a <style> element.  We don't want that here.  Yank it out.
    css_prefix = '<style type="text/css">\n\n'
    if not css.startswith(css_prefix):
        raise SeanoMarkupException('CSS returned from the reStructuredText compiler has an unexpected prefix: %s', css)
    css = css[len(css_prefix):]

    css_suffix = '\n\n</style>\n'
    if not css.endswith(css_suffix):
        raise SeanoMarkupException('CSS returned from the reStructuredText compiler has an unexpected suffix: %s', css)
    css = css[:len(css) - len(css_suffix)]

    # The default Pygments CSS does not work in dark mode; let's fix that:
    css = css + '''

@media (prefers-color-scheme: dark) {
    /* Custom overrides for Pygments so that it doesn't suck in dark mode */
    pre.code .ln { color: lightgrey; } /* line numbers */
    pre.code .comment, code .comment { color: rgb(127, 139, 151) }
    pre.code .keyword, code .keyword { color: rgb(236, 236, 22) }
    pre.code .literal.string, code .literal.string { color: rgb(217, 200, 124) }
    pre.code .name.builtin, code .name.builtin { color: rgb(255, 60, 255) }
}'''

    return SeanoHtmlFragment(html=html, css=css)


class SeanoMarkup(object):
    """
    Base class for encapsulating some rich text markup that can self-convert to
    various output formats.  Includes some stock shared code to support some of
    the common types that are, under the hoods, derivatives, rather than
    distinct types, allowing the real subclasses to not need to reimplement
    everything all the time.
    """

    def __init__(self, payload, localization=None, tags=None):
        self.payload = payload
        self.localization = localization
        self.tags = tags or []

    def deep_copy(self):
        return self.__class__(payload=self.payload, localization=self.localization, tags=self.tags)

    def __bool__(self): return bool(self.payload)
    def __nonzero__(self): return bool(self.payload)

    def __eq__(self, other): return \
        self.__class__ == other.__class__ and \
        self.payload == other.payload and \
        self.localization == other.localization and \
        self.tags == other.tags

    def __str__(self): return repr(self.payload)
    def __repr__(self): return str(self)

    _line_pattern = re.compile(r'''^<p[^>]*>(?P<contents>.*)</p>\s*$''', re.MULTILINE | re.DOTALL)
    def toHtmlLine(self):
        '''
        Returns the same as `toHtmlBlock()`, except sans a top-level element.

        On soft errors, this function may return an empty `SeanoHtmlFragment`
        object.  This function never returns `None`.

        The algorithm that strips the top-level element off of the return value
        is fairly stupid; try to not pass unknown garbage into this function.
        '''
        result = self.toHtmlBlock()
        if not result:
            return result
        m = self._line_pattern.match(result.html)
        if not m:
            raise SeanoMarkupException('Compiled HTML does not look like a single line: %s' % (result.html,))
        result.html = m.group('contents').strip()
        return result

    # ABK: toHtmlBlock() is the main function that subclasses need to implement


class SeanoPlainText(SeanoMarkup):
    def toHtmlLine(self):
        return SeanoHtmlFragment(html=html.escape(self.payload))

    def toHtmlBlock(self):
        result = self.toHtmlLine()
        result.html = '<p>' + result.html + '</p>'
        return result


class SeanoFalseyMarkup(SeanoMarkup):
    def __init__(self, explanation=''):
        self.explanation = explanation

    def __eq__(self, other): return \
        self.__class__ == other.__class__ and \
        self.explanation == other.explanation

    def deep_copy(self): return SeanoFalseyMarkup(explanation=self.explanation)
    def __bool__(self): return False
    def __nonzero__(self): return False
    def __str__(self): return '<FalseyMarkup>'
    def toHtmlLine(self): return SeanoHtmlFragment(html='')
    def toHtmlBlock(self): return SeanoHtmlFragment(html='')


class SeanoReStructuredText(SeanoMarkup):
    def __str__(self): return 'RST(%s)' % (super().__str__())

    def toHtmlBlock(self):
        return _seano_rst_to_some_html(self.payload,
                                      writer_class=docutils.writers.html4css1.Writer,
                                      translator_class=SeanoSingleFileHtmlTranslator)


class SeanoMarkdown(SeanoMarkup):
    def __str__(self): return 'MD(%s)' % (super().__str__())

    def toHtmlBlock(self):
        return SeanoHtmlFragment(
            html=markdown.markdown(
                self.payload,
                extensions=['pymdownx.superfences', 'pymdownx.highlight'],
            ),
        )


SUPPORTED_MARKUP = {
    'rst': SeanoReStructuredText,
    'md': SeanoMarkdown,
    '': SeanoPlainText,
}



"""
support/seano/views/shared/components.py

Shared code that renders certain common view components.
"""
import itertools
from .hlist import SeanoUnlocalizedHListIterNode
from .links import get_ticket_display_name


def seano_render_html_ticket(url):
    ticket = '<a href="%s" target="_blank">%s</a>' % (url, get_ticket_display_name(url))
    return '<span style="font-size:75%">' + ticket + '</span>'


def seano_html_hlist_blob_formatter_simple(node):
    return node.element.toHtmlBlock().html


def seano_html_hlist_line_formatter_simple(node):
    return node.element.toHtmlLine().html


def seano_html_hlist_line_formatter_text_with_tickets(notes, line_formatter=None):  #pylint: disable=C0103
    line_formatter = line_formatter or seano_html_hlist_line_formatter_simple
    def formatter(node):
        result = line_formatter(node)

        # To minimize UX noise, we don't want to print tickets on literally every line.
        # As an easy solution, only print tickets on the first line in a note tree where
        # that line and all sub-lines have the same ticket list.

        # Has one of our parents already printed?
        if getattr(node, 'is_printed', False):
            # One of our parents was already printed.  Bail.
            return result

        # None of our parents have printed.  Do all of our children have the
        # same tickets as us?

        def get_tickets(note_ids):
            return itertools.chain(*[note.get('tickets') or [] for note in notes if note['id'] in note_ids])
        tickets = list(get_tickets(note_ids=node.element.tags))
        tickets_set = set(tickets)

        for child in node.walk():
            if tickets_set != set(get_tickets(child.note_ids)):
                # One of our children has a different note set than us.
                # Defer printing of tickets until we get to a deeper sub-line.
                return result

        # We've decided to print tickets on this line.  Silence tickets on all sub-lines:
        def silence(x):
            x.is_printed = True
            for child in x.children:
                silence(child)
        silence(node)

        # Remove duplicate tickets without changing sort order:
        def dedup(lst):
            seen = set()
            for x in lst:
                if x not in seen:
                    seen.add(x)
                    yield x
        tickets = dedup(tickets)

        # Compile tickets into HTML:
        tickets = [seano_render_html_ticket(x) for x in tickets]

        # Return entire note line, with all of the tickets:
        return ' '.join([result] + tickets)

    return formatter


def seano_render_html_hlist(hlist, is_blob_field=False, block_formatter=None, line_formatter=None):

    block_formatter = block_formatter or seano_html_hlist_blob_formatter_simple
    line_formatter = line_formatter or seano_html_hlist_line_formatter_simple

    def _run(node):

        if node.level > 0:
            yield '<li>'
            yield line_formatter(node=node)
        elif node.element:
            yield block_formatter(node=node)

        if node.children:
            if node.level >= 0:
                yield '<ul>'
            for child in node.children:
                for hunk in _run(node=child):
                    yield hunk
            if node.level >= 0:
                yield '</ul>'

        if node.level > 0:
            yield '</li>'

    return ''.join(_run(node=SeanoUnlocalizedHListIterNode(node=hlist, level=-1 if is_blob_field else 0)))
